from pathlib import Path
from typer import Argument, Option, Typer
import re
import sys

from . import app, DRY_RUN, expand_files, split_file


@app.command()
def rename(
    files: list[Path],
    dry_run: bool = Option(
        DRY_RUN,
        "--dry-run",
        "-d",
        help="Print commands, don't execute them",
    ),
    force: bool = Option(
        False,
        "--force",
        "-f",
        help="Overwrite existing files",
    ),
) -> None:
    missing, zoia, not_zoia, already_exists = [], [], [], []
    for file in expand_files(files):
        try:
            prefix, base = split_file(file)
        except ValueError:
            not_zoia.append(file)
            continue
        if not file.exists():
            missing.append(file)
        else:
            new_file = file.parent / base
            if not force and new_file.exists():
                already_exists.append(new_file)
            else:
                zoia.append((file, new_file))

    if not_zoia:
        print(f"WARNING: not zoia files: {not_zoia}", file=sys.stderr)

    errors = []
    if missing:
        errors.append(f"Missing: {missing}")
    if already_exists:
        errors.append(f"Cannot overwrite: {already_exists}. Use -f to replace")
    if not (errors or zoia):
        errors.append("No files to rename")

    if errors:
        sys.exit('\nERROR: '.join(("", *errors)).strip())

    for old, new in zoia:
        print(old, "->", new)
        if not dry_run:
            old.rename(new)



from typer import Typer
import re
import typing as t
from pathlib import Path


ZFILE_MATCH = re.compile(r"(\d\d\d_zoia_)(.*\.bin)").match
LIBFILE_MATCH = re.compile(r".*\.bin").match
DRY_RUN = False


app = Typer(
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
)


def split_file(file: str | Path) -> tuple[str, str]:
    s = Path(file).name
    if m := ZFILE_MATCH(s):
        return m.groups()
    if m := LIBFILE_MATCH(s):
        return "", *m.groups()
    raise ValueError("Not a patch file")



def expand_files(files: t.Iterable[Path]) -> t.Iterator[Path]:
    for f in files:
        if f.suffix == ".txt":
            li = f.read_text().splitlines()
            yield from expand_files(t for s in li if (t := s.partition("#")[0].strip()))
        else:
            yield f



#!/usr/bin/python3
"""Script for renaming binary files to contain a preferred numerical order.

This is for use with patch files for the Empress Zoia, which requires the patch files to
start with a three-digit number and doesn't permit duplicates.

Instead of manually renaming, you can create a PLAN file with:

    $ python3 repatch.py

Example output:

    Plan file saved to /home/you/to-zoia/PATCH-PLAN-2025-03-07-20-38-18.txt.

    You can reorder the items in that file by editing it, and then execute the plan by re-running
    this script with:
    repatch.py -p PATCH-PLAN-2025-03-07-20-38-18.txt -x
"""

import argparse
import datetime
import os
import re
import shutil
import sys
from typing import Optional

# Constants
MAX_ENTRIES = 64
NOW = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
BINARY_REGEX = re.compile(r".*\.bin")
PREFIX_REGEX = re.compile(r"^(- )?.?(\d+)(.*)")
BLANK_PATCH_NAME_SUFFIX = "_zoia_.bin"
BLANK_ZOIA_PATCH = b"\t" + (b"\x00" * 32767)


def parse_arguments():
    parser = argparse.ArgumentParser(description="File Reordering Tool")
    parser.add_argument(
        "-x",
        "--execute",
        action="store_true",
        help="Execute the renaming (non-dry run).",
    )

    def _os_path(string):
        if os.path.exists(string):
            return string
        else:
            raise NotADirectoryError(string)

    parser.add_argument(
        "-p",
        "--plan",
        type=_os_path,
        help="Path to a file with the binaries listed in the order they should go.",
    )
    return parser.parse_args()


def get_local_binary_files(path: str) -> list[str]:
    """Get a list of all files in the folder."""
    files = os.listdir(path)
    matching_files = []
    added_files = set()
    for file_name in sorted(files):
        match = BINARY_REGEX.match(file_name)
        if match and file_name not in added_files:
            matching_files.append(file_name)
            added_files.add(file_name)
    return matching_files


def matches_some_file(name_fragment: str, local_filenames: set[str]) -> str:
    """If the name-fragment is a fragment of some file in the local directory, return its name.

    Throws ValueError if there's no match or more than one match.
    """
    matches = []
    for filename in local_filenames:
        if name_fragment in filename:
            matches.append(filename)
    if len(matches) > 1:
        print(f"Multiple files matching the plan element {name_fragment} were found: {matches}")
        sys.exit(2)
    if not matches:
        print(f"Some files in the plan-file no longer exist (ex: {name_fragment}). Aborting.")
        sys.exit(3)
    return matches[0]


def generate_new_names(plan: list[str], local_binary_files: set[str]) -> list[tuple[Optional[str], str]]:
    """Process the plan text and return the old and new filenames.

    Note that the filenames in the plan are prefaced by "- " and followed by a newline.
    """
    ret = []
    for i in range(MAX_ENTRIES):
        new_num_prefix = str(i).zfill(3)
        if i < len(plan):
            file_name = plan[i]
            input_name = file_name.lstrip("- 0123456789_").strip()
            new_name = f"{new_num_prefix}_{input_name}"
            match = matches_some_file(input_name, local_binary_files)
        else:
            # We will write a blank patch so that the user can save their own patches.
            match = None
            new_name = str(i).zfill(3) + BLANK_PATCH_NAME_SUFFIX
        ret.append((match, new_name))
    return ret


def main():
    args = parse_arguments()
    folder_path = os.getcwd()
    #          | no plan file                      | plan file                                   |
    # dry run  | generate a plan file              | describe what we would do                   |
    # live run | describe and auto-move the files  | rename the files according to the plan file |

    # if dry run and no plan file, we should generate a plan file
    local_binary_files = get_local_binary_files(folder_path)
    if args.plan:
        with open(args.plan) as infile:
            plan = infile.readlines()
    else:
        if args.execute:
            # execute and not plan? weird, but not disallowed
            plan = local_binary_files
        # if not execute, then write a plan file for the user to edit
        else:
            plan_file_name = f"PATCH-PLAN-{NOW}.txt"
            plan_file_path = os.path.join(folder_path, plan_file_name)
            with open(plan_file_path, "w") as outfile:
                for filename in local_binary_files:
                    outfile.write(f"- {filename}\n")
            print(f"Plan file saved to {plan_file_path}.")
            print(
                "You can reorder the items in that file by editing it, and then execute the plan by re-running this script with:"
            )
            print(f"  repatch.py -p {plan_file_name} -x")
    # if dry run and plan file, we should output the new names.
    # if live run and no plan file, we should describe and auto-move the files
    # if live run and plan file, we should rename the files according to the plan file.
    if args.execute or args.plan:
        target_folder_path = f"repatched-{NOW}"
        old_and_new_names = generate_new_names(plan, local_binary_files)
        if args.execute:
            # ensure target directory exists.
            target_dir: str = os.path.join(folder_path, target_folder_path)
            if not os.path.exists(target_dir):
                os.makedirs(target_dir)

            print("Moving the files to the following paths:")
        else:
            print("Would move the files to the following paths:")
        for old_name, new_name in old_and_new_names:
            print(f"- {os.path.join(target_folder_path, new_name)}")
            if args.execute:
                # Check if we need to write a blank file.
                if old_name is None:
                    with open(os.path.join(target_dir, new_name), "wb") as outfile:
                        outfile.write(BLANK_ZOIA_PATCH)
                else:
                    shutil.copyfile(
                        os.path.join(folder_path, old_name),
                        os.path.join(target_dir, new_name),
                    )
        if not args.execute:
            print("Re-run again with -x / --execute to execute this plan.")


if __name__ == "__main__":
    main()



import collections
import datetime
import typing as t
from pathlib import Path
from typer import Argument, Option, Typer
import re
import shutil
import sys
import tomlkit

from . import app, DRY_RUN, expand_files, split_file

EMPTY_PATCH = Path(__file__).parents[1] / 'zoia_empty.bin'
TIMESTAMP = datetime.datetime.now().strftime("%d-%m-%Y_%H-%M-%S")

"""
TODO:

* write some tests
* write documentation

"""


@app.command()
def prepare(
    files: list[Path] = Argument(
        help="List of file names to prepare",
    ),
    output: Path = Option(
        Path(f"zoia-{TIMESTAMP}"),
        "--output",
        "-o",
        help="Directory to write patch files to",
    ),
    slots_file: Path = Option(
        Path("slots_file.toml"),
        "--slots-file",
        "-f",
        help="Set the file used to store slot lists",
    ),
    write_slots_file: bool = Option(
        True,
        "--write-slots-file",
        "-w/-n",
        help="If true, overwrite the slots file ",
    ),
    verbose: bool = Option(
        False,
        "--verbose",
        "-v",
        help="Print more information",
    ),
    dry_run: bool = Option(
        DRY_RUN,
        "--dry-run",
        "-d",
        help="Print commands, don't execute them",
    ),
    slot_count: int = Option(
        0,
        "--slot-count",
        "-s",
        help="How many slots to fill. If zero, fill to the last slot",
    ),
) -> None:
    files = list(expand_files(files))
    if slots_file.exists():
        cfg = tomlkit.loads(slots_file.read_text())
    else:
        cfg = tomlkit.TOMLDocument()

    if verbose and not output.exists():
        print("Making output directory", output)
    if not dry_run:
        output.mkdir(exist_ok=True, parents=True)

    # return [(f, f"{i:03}_zoia_{base}") for i, (f, base) in enumerate(slot_list)]

    slot_list = _compute_slot_list(cfg.get("slots", {}), files, slot_count)
    for i, (source, base) in enumerate(slot_list):
        assert isinstance(base, str), base
        target = output / f"{i:03}_zoia_{base}"
        if verbose:
            print(f"Copying {source} to {target}")
        if not dry_run:
            if write_slots_file and base:
                slot = cfg.setdefault("slots", {}).setdefault(f"{i:03}", [])
                if base not in slot:
                    slot.append(base)

            shutil.copy(str(source), str(target))

    if write_slots_file and not dry_run and cfg:
        slots_file.write_text(cfg.as_string())


def _compute_slot_list(
    slots: dict[str, t.Any],
    files: list[Path],
    slot_count: int,
):
    slot_assignments = {}
    todo, bad, bad_slots, collisions = [], [], [], []
    max_slot = max(len(files), slot_count) - 1
    for f in files:
        body, _, slot_str = f.name.partition(":")  # : indicates slot
        try:
            base = split_file(body)[1]
        except ValueError:
            bad.append(f)
            continue
        if not slot_str:
            todo.append((f, base))
            continue
        try:
            slot = int(slot_str, 10)
        except Exception:
            bad_slots.append(f)
            continue
        max_slot = max(slot, max_slot)
        if slot in slot_assignments:
            collisions.append(slot)
        else:
            slot_assignments[slot] = (f, base)

    errors = []
    if bad:
        errors.append(f"Not zoia: {bad}")
    if bad_slots:
        errors.append(f"Bad slot identifier: {bad_slots}")
    if collisions:
        errors.append(f"Slot numbers must be distinct: {collisions}")
    d = collections.Counter(base for f, base in todo)
    if dupes := [k for k, v in d.items() if v > 2]:
        errors.append(f"Duplicates: {dupes}")
    if errors:
        sys.exit('\nERROR: '.join(("", *errors)).strip())

    slot_list = [slot_assignments.get(i) for i in range(max_slot + 1)]

    todo_names = {base: f for f, base in todo}
    for i, s in enumerate(slot_list):
        if not todo_names:
            break
        if s:
            continue
        for sl in slots.get(f"{i:03}", ()):
            if f := todo_names.pop(sl, None):
                slot_list[i] = (f, sl)
                break

    names = [(f, b) for b, f in reversed(todo_names.items())]
    slot_list = [i or (names and names.pop()) or (EMPTY_PATCH, ".bin") for i in slot_list]
    assert not names, names

    while len(slot_list) > slot_count and not slot_list[-1]:
        slot_list.pop()

    return slot_list
